# Copyright 2012-2014 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

# Some code taken and modified from trac / GPL.

__metaclass__ = type

import collections
from ConfigParser import ConfigParser
from contextlib import contextmanager
import copy
import datetime
from dateutil.parser import parse
import logging.config
import os
from pyelasticsearch.exceptions import ElasticHttpNotFoundError
import re
import time
from urllib import unquote
import urlparse

# This is the base directory the project (i.e., the directory in which the
# setup.py and Makefile exist), the directory above which there are no project
# files.
project_root = os.path.dirname(os.path.dirname(__file__))


# Required for get_session_url on Precise, but not on Quantal.
urlparse.uses_query.append('mongodb')


_units = (
    (3600 * 24 * 365, '%d year'),
    (3600 * 24 * 30, '%d month'),
    (3600 * 24 * 7, '%d week'),
    (3600 * 24, '%d day'),
    (3600, '%d hour'),
    (60, '%d minute'),
)

# The MongoDB ID of the time the last ingest job finished in the 'status'
# collection.
LAST_INGEST_JOB_FINISHED = 'last_ingest_job_finished'
# The MongoDB ID of the time the server was started in the 'status' collection.
SERVER_STARTED = 'server_started'


def prettify_timedelta(diff):
    """Return a prettified time delta, suitable for pretty printing."""
    age_s = int(diff.days * 86400 + diff.seconds)

    for u, template in _units:
        r = float(age_s) / float(u)
        if r >= 1.9:
            r = int(round(r))
            v = template % r
            if r > 1:
                v += "s"
            return v
    return ''


def pretty_timedelta(time1, time2=None):
    """Calculate time delta between two `datetime` objects.
    (the result is somewhat imprecise, only use for prettyprinting).
    """
    if isinstance(time1, (str, unicode)):
        time1 = parse(time1)
    if time2 is None:
        time2 = datetime.datetime.utcnow()

    if time1 > time2:
        raise ValueError('The start date is before the end date.')

    diff = time2 - time1
    return prettify_timedelta(diff)


def configure_logging():
    logging.config.fileConfig(get_ini_path())


def get_ini_path():
    """Determine the absolute path of the selected ini file.

    The default ini file is the 'charmworld.ini', however it can be overridden
    by providing a INI=test.ini ENV variable.  Filenames are interpreted
    relative to the charmworld tree root regardless of the current working
    directory.
    """
    ini_file = os.environ.get('INI', 'charmworld.ini')
    ini_path = os.path.join(
        os.path.dirname(os.path.dirname(__file__)), ini_file)
    return ini_path


def get_ini():
    """Fetch the ini file in use and provide a dict of values back.

    This is to mimic the access provided to the Pyramid app to outside code.
    It provides a dict from the ConfigParser instance in order to mimic access
    of request.registry.settings. It only provides the main app section of
    config data.

    The default ini file is the 'charmworld.ini', however it can be overridden
    by providing a INI=test.ini ENV variable.   See get_ini_path.

    """
    ini = ConfigParser()
    with open(get_ini_path()) as fp:
        ini.readfp(fp)
    here = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../'))
    ini.set('app:main', 'here', here)
    return dict(ini.items("app:main"))


def get_session_url(settings):
    parsed = urlparse.urlparse(settings['mongo.url'])
    path = settings['mongo.database'] + '.sessions'
    return urlparse.urlunparse((parsed[0], parsed[1], path, parsed[3],
                                parsed[4], parsed[5]))


def quote_key(key):
    """Replace the symbols '.', '$', '%' with escape sequences.

    PyMongo version 2.4.2 for example does not allow dots in keys, and
    while older versions allowed to store dictionaries like {'a.b.c': 'd'},
    a query like collection.find({'a.b.c': 'd'}) would not find this
    record because the query searches for a record like
    {'a': {'b': {'c': 'd'}}}.

    Moreover, the '$' is reserved for query keywords so it should be
    avoided too.
    """
    for char, replacement in (('%', '%25'), ('$', '%24'), ('.', '%2E')):
        key = key.replace(char, replacement)
    return key


def _process_dict_keys(data, key_mapper):
    """Apply the key_mapper function to keys in the data dictionary.

    The dictionary `data` is modified in place.
    If the dictionary contains any dicts or lists, the function is called
    recursively on those inner structures with no limits.
    """
    if isinstance(data, collections.Mapping):
        tmp = {}
        for key, value in data.items():
            if isinstance(key, str) or isinstance(key, unicode):
                new_key = key_mapper(key)
                if new_key != key:
                    tmp[new_key] = value
                    del data[key]
            _process_dict_keys(value, key_mapper)
        data.update(tmp)
    elif (isinstance(data, collections.Sequence)
          and not isinstance(data, str) and not isinstance(data, unicode)):
        for item in data:
            _process_dict_keys(item, key_mapper)


def quote_yaml(orig_data):
    """Traverse the given object and its sub-objects; if the object
    is a dictionary, replace string keys containing at least one of
    '.', '$', '%' with values that can safely be stored in Mongo.
    """
    data = copy.deepcopy(orig_data)
    _process_dict_keys(data, quote_key)
    return data


def unquote_yaml(orig_data):
    """Unquote dictionary keys that were quoted by quote_yaml().
    """
    data = copy.deepcopy(orig_data)
    _process_dict_keys(data, unquote)
    return data


def timestamp(date):
    """Return an RFC 3339 timestamp representing the supplied date(time).

    datetimes are expected to be naive, and represent UTC time.
    """
    return date.isoformat() + 'Z'


class Timer:
    """Context manager for timing a block of code."""

    def __init__(self):
        self.start = None
        self.end = None

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end = time.time()

    def duration(self):
        return self.end - self.start


@contextmanager
def write_locked(lockable):
    lockable.lock_write()
    try:
        yield lockable
    finally:
        lockable.unlock()


@contextmanager
def read_locked(lockable):
    lockable.lock_read()
    try:
        yield lockable
    finally:
        lockable.unlock()


def utc_today():
    """The current date in UTC."""
    return datetime.datetime.utcnow().date()


# This regex identifies the version component of a bundle/charm ID to be
# removed by build_metric_key.  The version component starts with a forward
# slash, then has 1 or more digits, and then ends with a slash (which should
# not be removed, hence the "?=" lookahead assertion).
id_revision_regex = re.compile(r'/\d+(?=/)')


def build_metric_key(metric_name, item_id):
    """Construct a metric key from a metric name and an item ID.

    The item (charm/bundle) ID may contain a version which will be disregarded
    when constructing the metric key.
    """
    # Metrics are stored for the versionless ID.
    versionless_id = id_revision_regex.sub('', item_id)
    return '/'.join([metric_name, versionless_id])


def set_status_timestamp(database, id):
    database.server_status.save({
        '_id': id,
        'time': time.time(),
    })


def record_server_start_time(database):
    hostname = os.uname()[1]
    database.server_status.update(
        {
            '_id': SERVER_STARTED,
        },
        {
            '$set': {
                hostname: time.time(),
            },
        },
        upsert=True)


def delete_charm(data, db, index_client):
    """Completely remove the charm from the system.

    That includes unfeaturing it and removing it from Elasticsearch and
    MongoDB.
    """
    from charmworld.models import (
        CharmSource,
        FeaturedSource,
    )
    from charmworld.search import CHARM
    # The charm needs to be removed from the search engine.
    try:
        index_client.delete_charm(data)
    except ElasticHttpNotFoundError:
        # It is fine if the charm was not indexed.
        pass

    # If the charm is featured, we want to unfeature it.
    source = FeaturedSource.from_db(db)
    source.clear_featured(data, CHARM)

    # Finally, remove all the revisions of the charm document from the
    # database.
    charm_query = CharmSource.make_all_revisions_query(data)
    db.charms.remove(charm_query)


def delete_bundle(data, db, index_client):
    from charmworld.models import (
        Bundle,
        FeaturedSource,
    )
    from charmworld.search import BUNDLE
    try:
        index_client.delete_bundle(data)
    except ElasticHttpNotFoundError:
        # It is fine if the bundle was not indexed.
        pass

    # Clear if featured.
    source = FeaturedSource.from_db(db)
    source.clear_featured(data, BUNDLE)

    # Remove from mongodb.
    bundle = Bundle(data)
    db.bundles.remove({'_id': bundle.id})
