# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type


from datetime import datetime

from charmworld.models import (
    Charm,
    CharmSource,
    QADataSource,
)
from charmworld.views.tools import (
    charm_review,
    format_items_for_queue,
    get_missing_qa_items,
    get_sparklines,
)
from charmworld.testing import factory
from charmworld.testing import ViewTestBase
from charmworld.testing import WebTestBase
from charmworld.testing.html import parse


class TestToolsHome(WebTestBase):

    def test_tools_view(self):
        """Make sure the tools view renders correctly."""
        resp = self.app.get(
            '/tools',
            status=200)

        body = resp.body
        self.assertIn("Charm Tools", body)


class TestReviewDisplay(WebTestBase):

    def test_last_modified_formatting(self):
        self.db.review_queue.save(
            factory.make_review_entry(date_modified=datetime(2012, 1, 2)))
        self.db.review_queue.save(
            factory.make_review_entry(date_modified=datetime(2012, 1, 3)))
        resp = self.app.get('/tools/review-queue')
        tree = parse(resp.body)
        headings = [th.text for th in tree.findall('body//table/thead/tr/th')]
        self.assertEqual(
            ['Age', 'Date Created', 'Last Modified', 'Summary', 'Status'],
            headings)
        rows = tree.findall('body//table/tr')
        self.assertEqual(2, len(rows))
        colmap = {}
        for row in rows:
            columns = row.findall('td')
            self.assertEqual(len(headings), len(columns))
            [colmap.setdefault(heading, []).append(col.text)
             for heading, col in zip(headings, columns)]
        self.assertEqual(['2012/01/02', '2012/01/03'], colmap['Last Modified'])


class TestReviews(ViewTestBase):

    def test_missing_qa_data_in_queue(self):
        client = self.use_index_client()
        ignore, charm = factory.makeCharm(self.db, promulgated=True)
        client.index_charm(charm)
        qa = get_missing_qa_items(
            CharmSource.from_request(self.getRequest()), self.db)
        self.assertEqual(1, len(qa))

    def test_charm_with_qa_data_matches_name_and_series(self):
        # Only one of the two charms should make it into the queue.
        client = self.use_index_client()
        _, charm_json = factory.makeCharm(self.db, promulgated=True)
        client.index_charm(charm_json)
        charm = Charm(charm_json)
        _, charm_json_2 = factory.makeCharm(
            self.db,
            name=charm.name,
            series='a-different-series',
            promulgated=True
        )
        client.index_charm(charm_json_2)
        qa_source = QADataSource.from_db(self.db)
        qa_source.save_qa_data(charm, [])
        qa = get_missing_qa_items(
            CharmSource.from_request(self.getRequest()), self.db)
        self.assertEqual(1, len(qa))

    def test_charm_with_qa_data_not_in_queue(self):
        client = self.use_index_client()
        _, charm_json = factory.makeCharm(self.db, promulgated=True)
        client.index_charm(charm_json)
        charm = Charm(charm_json)
        qa_source = QADataSource.from_db(self.db)
        qa_source.save_qa_data(charm, [])
        qa = get_missing_qa_items(
            CharmSource.from_request(self.getRequest()), self.db)
        self.assertEqual(0, len(qa))

    def test_unpromulgated_with_missing_qa_not_in_queue(self):
        client = self.use_index_client()
        _, charm_json = factory.makeCharm(self.db, promulgated=False)
        client.index_charm(charm_json)
        qa = get_missing_qa_items(
            CharmSource.from_request(self.getRequest()), self.db)
        self.assertEqual(0, len(qa))

    def test_askubuntu_questions_in_review(self):
        entry = factory.make_review_entry()
        entry['type'] = 'question'
        self.db.answer_queue.insert([entry])
        resp = charm_review(self.getRequest())
        q = resp['queue']
        self.assertEqual(1, len(q))
        self.assertEqual('question', q[0]['type'])

    def test_github_pull_requests_in_review(self):
        entry = factory.make_review_entry()
        entry['type'] = 'pull request'
        self.db.github_pull_requests.insert([entry])
        resp = charm_review(self.getRequest())
        q = resp['queue']
        self.assertEqual(1, len(q))
        self.assertEqual('pull request', q[0]['type'])

    def test_qa_entry_formatting(self):
        ignore, charm = factory.makeCharm(self.db)
        charm = Charm(charm)
        entry = format_items_for_queue([charm], 'qa')[0]
        self.assertEqual('qa', entry['type'])
        self.assertEqual(entry['item'], "%s/qa/edit" % charm.short_url)
        self.assertEqual(
            entry['summary'], 'Missing QA data on %s' % charm.name)

    def test_proof_entry_formatting(self):
        ignore, charm = factory.makeCharm(self.db)
        charm = Charm(charm)
        entry = format_items_for_queue([charm], 'proof')[0]
        self.assertEqual('proof', entry['type'])
        self.assertEqual(entry['item'], "%s#proof" % charm.short_url)
        self.assertEqual(
            entry['summary'], 'Proof error on %s' % charm.name)

    def test_store_entry_formatting(self):
        ignore, charm = factory.makeCharm(self.db)
        charm = Charm(charm)
        entry = format_items_for_queue([charm], 'store')[0]
        self.assertEqual('store', entry['type'])
        self.assertEqual(
            entry['item'], 'http://code.launchpad.net/%s' % charm.branch_spec)
        self.assertEqual(
            entry['summary'], 'Store error on %s' % charm.name)

    def test_empty_sparklines(self):
        sparklines = get_sparklines([])
        self.assertEqual('', sparklines['means'])

    def test_single_sparklines(self):
        sparklines = get_sparklines([{'mean': 3}])
        self.assertEqual('3', sparklines['means'])

    def test_sparklines(self):
        sparklines = get_sparklines([
            {'mean': 1},
            {'mean': 3}])
        self.assertEqual('1,3', sparklines['means'])

    def test_current_average_is_most_recent(self):
        #self.db.review_queue.save(
            #factory.make_review_entry(date_modified=datetime(2012, 1, 2)))
        self.db.review_queue_latency.save(
            {'mean': 3600 * 24, 'date': datetime(2012, 1, 2)})
        self.db.review_queue_latency.save(
            {'mean': 3600 * 2 * 24, 'date': datetime(2013, 1, 2)})
        resp = charm_review(self.getRequest())
        self.assertEqual('2 days', resp['mean'])
